import { CancellationTokenSource, assert } from "@surface/core";
import { tryEvaluateExpressionByTraceable, tryEvaluateKeyExpressionByTraceable, tryEvaluatePatternByTraceable, tryObserveByObservable, tryObserveKeyByObservable, } from "../common.js";
import TemplateMetadata from "../metadata/template-metadata.js";
import TemplateProcessor from "../processors/template-processor.js";
import { scheduler } from "../singletons.js";
import TemplateBlock from "./template-block.js";
export default class PlaceholderDirective {
    constructor(template, descriptor, context) {
        this.cancellationTokenSource = new CancellationTokenSource();
        this.templateBlock = new TemplateBlock();
        this.currentDisposable = null;
        this.disposed = false;
        this.key = "";
        this.lazyInjectionCancellationTokenSource = new CancellationTokenSource();
        this.subscription = null;
        this.applyInjection = () => {
            const injection = this.metadata.injections.get(this.key);
            this.inject(injection);
        };
        this.applyLazyInjection = () => {
            this.lazyInjectionCancellationTokenSource = new CancellationTokenSource();
            void scheduler.enqueue(this.applyInjection, "low", this.lazyInjectionCancellationTokenSource.token);
        };
        this.inject = (injection) => {
            this.lazyInjectionCancellationTokenSource.cancel();
            this.currentDisposable?.dispose();
            this.currentDisposable = null;
            this.subscription?.unsubscribe();
            this.subscription = null;
            const task = (this.injectionContext = injection)
                ? this.task
                : this.defaultTask;
            const listener = () => void scheduler.enqueue(task, "normal", this.cancellationTokenSource.token);
            this.subscription = tryObserveByObservable(this.context.scope, this.descriptor, listener, true);
            listener();
        };
        this.onKeyChange = () => {
            if (this.key) {
                this.metadata.defaults.delete(this.key);
                this.metadata.placeholders.delete(this.key);
            }
            this.key = `${tryEvaluateKeyExpressionByTraceable(this.context.scope, this.descriptor)}`;
            this.metadata.defaults.set(this.key, this.applyLazyInjection);
            this.metadata.placeholders.set(this.key, this.inject);
            if (this.context.host.isConnected) {
                this.applyInjection();
            }
            else {
                this.applyLazyInjection();
            }
        };
        this.task = () => {
            assert(this.injectionContext);
            this.currentDisposable?.dispose();
            this.templateBlock.clear();
            const value = tryEvaluateExpressionByTraceable(this.context.scope, this.descriptor);
            const directiveScope = tryEvaluatePatternByTraceable(this.context.scope, value, this.injectionContext.descriptor);
            const mergedScope = { ...this.injectionContext.scope, ...directiveScope };
            const content = this.injectionContext.template.content.cloneNode(true);
            const context = {
                directives: this.injectionContext.customDirectives,
                host: this.injectionContext.host,
                parentNode: this.injectionContext.parentNode,
                root: content,
                scope: mergedScope,
                templateDescriptor: this.injectionContext.descriptor.descriptor,
            };
            const disposable = TemplateProcessor.process(context);
            this.templateBlock.setContent(content);
            this.currentDisposable = disposable;
        };
        this.defaultTask = () => {
            this.currentDisposable?.dispose();
            this.templateBlock.clear();
            const content = this.template.content.cloneNode(true);
            const context = {
                directives: this.context.directives,
                host: this.context.host,
                parentNode: this.context.parentNode,
                root: content,
                scope: this.context.scope,
                templateDescriptor: this.descriptor.descriptor,
            };
            const disposable = TemplateProcessor.process(context);
            this.templateBlock.setContent(content);
            this.currentDisposable = disposable;
        };
        this.template = template;
        this.descriptor = descriptor;
        this.context = context;
        this.metadata = TemplateMetadata.from(context.host);
        assert(template.parentNode);
        const parent = template.parentNode;
        this.templateBlock.insertAt(parent, template);
        this.keySubscription = tryObserveKeyByObservable(context.scope, descriptor, this.onKeyChange, true);
        this.onKeyChange();
    }
    dispose() {
        if (!this.disposed) {
            this.cancellationTokenSource.cancel();
            this.currentDisposable?.dispose();
            this.keySubscription.unsubscribe();
            this.subscription.unsubscribe();
            this.metadata.defaults.delete(this.key);
            this.metadata.placeholders.delete(this.key);
            this.templateBlock.dispose();
            this.disposed = true;
        }
    }
}
